import {implementations, commonSpecs} from '@/core/Dispatcher'
import {joinTypes} from '@/core/type'
import {$reflect} from '@/interfaces/type'

import type {
   CommonInterface, CommonSignature, NaNType, OneType, ZeroType
} from '@/interfaces/type'

export type Complex<T> = {re: T, im: T}

export const Complex_type = {
   name: 'Complex',
   test: <T>(dep: {testT: (z: unknown) => z is T}) =>
      (z: unknown): z is Complex<T> =>
         typeof z === 'object' && z != null && 're' in z && 'im' in z
         && dep.testT(z.re) && dep.testT(z.im),
   infer: (dep: {typeOf: CommonSignature<undefined>['typeOf']}) =>
      (z: Complex<unknown>) => joinTypes(dep.typeOf(z.re), dep.typeOf(z.im)),
   from: {
      Complex: <U,T>(dep: {convert: CommonSignature<U,T>['convert']}) =>
         (z: Complex<U>) => ({re: dep.convert(z.re), im: dep.convert(z.im)}),
      T: <T>(dep: {zero: CommonSignature<T>['zero']}) =>
         (t: T) => ({re: t, im: dep.zero(t)})
   }
}

declare module "@/interfaces/type" {
   interface AssociatedTypes<T> {
      Complex: T extends Complex<infer R> ? {
         type: Complex<R>
         zero: Complex<ZeroType<R>>
         one: Complex<OneType<R> | ZeroType<R>>
         nan: Complex<NaNType<R>>
         real: RealType<R>
         closure: T
      } : never
   }

   interface CommonInterface<T, Aux> {
      complex: (re: T, im?: T) => Complex<T>
   }
}

// internal builder
const cplex = <T>(a:T, b:T): Complex<T> => ({re: a, im: b})

export function lift<T>() {
   return implementations<CommonSignature<T>>()
      .dependent({zero: {}}, {
         complex: dep => (a, b) => cplex(a, b || dep.zero(a))
      }).ship()
}

export function common<T>() {
   const baseSignature = commonSpecs<T>()

   return implementations<CommonSignature<Complex<T>>>()
      .dependent(baseSignature({zero: {}}), { 
         zero: dep => z => cplex(dep.zero(z.re), dep.zero(z.re))
      })
      .dependent(baseSignature({zero: {}, one: {}}), {
         one: dep => z =>
            cplex<OneType<T> | ZeroType<T>>(dep.one(z.re), dep.zero(z.re))
      })
      .dependent(baseSignature({nan: {}}), {
         nan: dep => z => cplex(dep.nan(z.re), dep.nan(z.re))
      })
      .dependent(baseSignature({re: {}}), {
         re: dep => z => dep.re(z.re)
      })
      .ship()
}

$reflect!([common, lift])
